/**
 * MPQSyncTask.java: Subclass ofASyncTask, to synchronize MPQ data with an Internet server.
 *            Copyright (C) 2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2013-04-20 donfede@casagrau.org
 * credit: android training / connecting to the network -- https://developer.android.com/training/basics/network-ops/connecting.html
 * 		   android training / managing network usage -- https://developer.android.com/training/basics/network-ops/managing.html
 *
 * $Id: MPQSyncTask.java,v 1.10 2013-04-25 04:35:50 donfede Exp $
 */


package org.casagrau.mpq.mpq_android;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.protocol.HTTP;
import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.ContentValues;
import android.content.Context;
import android.content.SharedPreferences;
import android.database.Cursor;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import android.net.Uri;
import android.os.AsyncTask;
import android.preference.PreferenceManager;
import android.support.v4.app.NotificationCompat;
import android.telephony.TelephonyManager;
import android.util.Log;
import android.widget.Toast;

/**
 * AsyncTask subclass to connect the MPQ web server and sync database contents.
 */
////public class MPQSyncTask extends AsyncTask<Object, Integer, Integer> {
public class MPQSyncTask extends AsyncTask<Void, Integer, Integer> {
	private static final String LOGTAG = "MPQSyncTask";
	
	private static final int NOTIFICATION_SYNC_ID = R.string.notification_sservice;
	
	private final String SCRUD_OP = "scrud_op";			// Flag to denote CRUD Operation.
	private final String SCRUD_OBJECT = "scrud_object";	// Flag to denote CRUD Object.
	
	private final String SCRUD_OP_INSUPD = "scrud_op_insupd";	// Operation to insert or update.
	private final String SCRUD_OP_DEL = "scrud_op_del";			// Operation to delete.
	private final String SCRUD_OP_META = "scrud_op_meta";		// Operation to send meta data (user and device id).
	private final String SCRUD_OBJECT_GOALS = "scrud_object_goals";				// Object representing goals table.
	private final String SCRUD_OBJECT_INSTRUMENTS = "scrud_object_instruments";	// Object representing instruments table.
	private final String SCRUD_OBJECT_PRACTICES = "scrud_object_practices";		// Object representing practices table.
	
	private Context mContext;
	private NotificationManager mNotificationManager;

	public MPQSyncTask(Object o0) {
        Log.d(LOGTAG, "CONSTRUCTOR() 00  ");
		mContext = (Context) o0;
	}
	
	@Override
	protected void onPreExecute() {
        Log.d(LOGTAG, "onPreExecute() 00  ");
        
        /**
         * Check network status; report error and cancel() if there is a problem.
         */
        ConnectivityManager connMgr;
        connMgr = (ConnectivityManager) mContext.getSystemService(Context.CONNECTIVITY_SERVICE);
        NetworkInfo networkInfo = connMgr.getActiveNetworkInfo();

        boolean wifiConnected;
        boolean mobileConnected;
        if ( (networkInfo != null) && (networkInfo.isConnected()) ) {
            Log.d(LOGTAG, "onPreExecute() 41 Connected: ");
            wifiConnected = networkInfo.getType() == ConnectivityManager.TYPE_WIFI;
            mobileConnected = networkInfo.getType() == ConnectivityManager.TYPE_MOBILE;
        } else { // Display error and cancel further processing by AsyncTask.
            Log.e(LOGTAG, "ERROR: onPreExecute() 43 NOT connected -- CANCEL:");
    		Toast.makeText(mContext,"MPQ: ERROR - no network connection.  Enable WIFI, and try again.", Toast.LENGTH_LONG).show();
            wifiConnected = false;
            mobileConnected = false;
            cancel(true);
        }
        Log.d(LOGTAG, "onPreExecute() 47 ActiveNetwork Wifi connected: " + wifiConnected);
        Log.d(LOGTAG, "onPreExecute() 48 ActiveNetwork Mobile connected: " + mobileConnected);
        
        // Otherwise, display notification and continue with task.
        showNotification();
        Log.d(LOGTAG, "onPreExecute() 99");
	}

	
	@Override
////	protected Integer doInBackground(Object... arg0) {
	protected Integer doInBackground(Void... arg0) {
        Log.d(LOGTAG, "doInBackground() 00");

        Log.d(LOGTAG, "doInBackground() 10 cancelled:" + isCancelled());
        
        // connect to server
        String urlStr = mContext.getString(R.string.sync_recv_url);
        Log.d(LOGTAG, "doInBackground() 13 urlStr:" + urlStr);
        HttpClient httpclient = new DefaultHttpClient();
        HttpPost httppost = new HttpPost(urlStr);

        // Prepare chunks of data to transfer with server.
        // FIXME: keep or delete -- final int SYNC_CHUNK_SIZE = 10;
        
        // Select all since last sync (all if sync is null)
        
        // Initialize JSON object and 'array', along with other vars used.
        JSONObject ljObject;
        JSONArray ljArray = new JSONArray();
        StringEntity lentity;
        long   ldb_last_sync = 0;
        String ldb_device_id; 
        String ldb_userid;
        String lpref_userid;
        int    ldb_dbver;
        String ldb_mpqver;
        int    ldb_mpqver_int;
        Uri luri;
        Cursor lcursor;
        String [] lprojection;
        String lselection;
        int lrows;
		long lnewNowSec = java.lang.System.currentTimeMillis();
        
        // Get last sync date and device id from sync table ;) .
        luri = SessionsDB.CONTENT_SYNC_URI;
        lprojection = new String[] {SessionsDB.SYNC_DATE_STAMP, SessionsDB.SYNC_DEVICE, SessionsDB.SYNC_USERID, SessionsDB.SYNC_DBVER, SessionsDB.SYNC_MPQVER, SessionsDB.SYNC_MPQVER_INT};
        lselection = SessionsDB._ID + " = 1 ";
		lcursor = mContext.getContentResolver().query(luri, lprojection, null, null, null);
		Log.d(LOGTAG, "doInBackground() 16   cur: " + lcursor.toString() + " count: " + lcursor.getCount() );
		if (1 != lcursor.getCount()) {
            Log.e(LOGTAG, "ERROR: doInBackground() 16XXX bad count from Sync query.  Early exit." + lcursor.getCount());
			return -1;
		}
		ldb_last_sync = lcursor.getInt(0);
		ldb_device_id = lcursor.getString(1);
		ldb_userid = lcursor.getString(2);
		ldb_dbver = lcursor.getInt(3);
		ldb_mpqver = lcursor.getString(4);
		ldb_mpqver_int = lcursor.getInt(5);
		Log.w(LOGTAG, "doInBackground() 17   db sync: " + ldb_last_sync);
		Log.w(LOGTAG, "doInBackground() 18   db device: " + ldb_device_id);
		Log.d(LOGTAG, "doInBackground() 18AA   userid: " + ldb_userid + " dbver: " + ldb_dbver + " mpq ver: " + ldb_mpqver + " mpq verI: " + ldb_mpqver_int);		
		lcursor.close();

		// Check current device ID; update sync table if necessary.
        TelephonyManager telephonyManager = (TelephonyManager)mContext.getSystemService(Context.TELEPHONY_SERVICE);
        String ldid = getDeviceID(telephonyManager);
        Log.w(LOGTAG, "doInBackground() 21 dev id:" + ldid + " db_did: " + ldb_device_id);
        // FIXME: if db does not match actual device, then update.        
        if (!ldid.equals(ldb_device_id)) {
            Log.w(LOGTAG, "doInBackground() 22  updating device in sync db");
            ContentValues lval = new ContentValues();
            lval.put(SessionsDB.SYNC_DEVICE, ldid);
            luri = SessionsDB.CONTENT_SYNC_URI;
            lselection = SessionsDB._ID + " = 1 ";

            int lrc = mContext.getContentResolver().update(luri, lval, lselection, null);
            ldb_device_id = ldid;
            Log.w(LOGTAG, "doInBackground() 23  updating device in sync db RC: " + lrc);
        }
        // Check userid; update sync table if necessary.
		SharedPreferences lsharedPref = PreferenceManager.getDefaultSharedPreferences(mContext);
		lpref_userid = lsharedPref.getString(mContext.getString(R.string.pref_userid_key), mContext.getString(R.string.pref_userid_default));
		if (!lpref_userid.equals(ldb_userid)) {
            Log.w(LOGTAG, "doInBackground() 25  updating userid sync db");
            ContentValues lval = new ContentValues();
            lval.put(SessionsDB.SYNC_USERID, lpref_userid);
            luri = SessionsDB.CONTENT_SYNC_URI;
            lselection = SessionsDB._ID + " = 1 ";
            int lrc = mContext.getContentResolver().update(luri, lval, lselection, null);
            ldb_userid = lpref_userid;
            Log.w(LOGTAG, "doInBackground() 26  updating userid sync db RC: " + lrc);
        }
        
       
        // Get sync meta data -- add to JSON array
	    ljObject = new JSONObject();
	    try {
		    // table -- N/A for sync meta data
		    // operation
		    ljObject.put(SCRUD_OP, SCRUD_OP_META);
		    // data fields
		    ljObject.put(SessionsDB.SYNC_USERID, ldb_userid);
		    ljObject.put(SessionsDB.SYNC_DEVICE, ldb_device_id);
		    ljObject.put(SessionsDB.SYNC_DBVER, ldb_dbver);
		    ljObject.put(SessionsDB.SYNC_MPQVER, ldb_mpqver);
		    ljObject.put(SessionsDB.SYNC_MPQVER_INT, ldb_mpqver_int);
		} catch (JSONException e) {
            Log.e(LOGTAG, "ERROR: doInBackground() 26XX  JSON Exception SYNC -- exiting");
			e.printStackTrace();
		}
		Log.d(LOGTAG, "doInBackground() 27CC   ljobject: " + ljObject.toString());
	    ljArray.put(ljObject);
        
        // Get all Goals edited since last sync -- add to JSON array
        luri = SessionsDB.CONTENT_GOALS_URI;
        lprojection = new String[] {SessionsDB._ID, SessionsDB.GOAL_NAME, SessionsDB.GOAL_DESC, SessionsDB.GOAL_DATA_VER};
        lselection = SessionsDB.GOAL_DATA_VER + " >= " + ldb_last_sync;
		lcursor = mContext.getContentResolver().query(luri, lprojection, lselection, null, null);
		if (null != lcursor) {
			Log.d(LOGTAG, "doInBackground() 33AA   Goals Q cur: " + lcursor.toString() + " sel: " + lselection + " proj: " + lprojection);
			lrows = lcursor.getCount();
			Log.w(LOGTAG, "doInBackground() 33BB cursor lrows: " + lrows );
			lcursor.moveToFirst();
			
			for (int i = 0;i < lrows; i++) {
			    ljObject = new JSONObject();
			    
			    try {
				    // table
				    ljObject.put(SCRUD_OBJECT, SCRUD_OBJECT_GOALS);
				    // operation
				    ljObject.put(SCRUD_OP, SCRUD_OP_INSUPD);
				    // data fields
				    ljObject.put(SessionsDB._ID, lcursor.getInt(0));
				    ljObject.put(SessionsDB.GOAL_NAME, lcursor.getString(1));
					ljObject.put(SessionsDB.GOAL_DESC, lcursor.getString(2));
					ljObject.put(SessionsDB.GOAL_DATA_VER, lcursor.getString(3));
				} catch (JSONException e) {
		            Log.e(LOGTAG, "ERROR: doInBackground() 33XX  JSON Exception GOALS -- exiting");
					e.printStackTrace();
				}
				Log.d(LOGTAG, "doInBackground() 33YY   ljobject: " + ljObject.toString());
			    ljArray.put(ljObject);
			    
				lcursor.moveToNext();
			} // END for()
		} else {
			Log.d(LOGTAG, "doInBackground() 34AA   Goals Q cur: " + lcursor + " sel: " + lselection + " proj: " + lprojection);
		}
		Log.d(LOGTAG, "doInBackground() 34ZZ   len: " +  ljArray.length() + " jarrayStr: " + ljArray.toString());

        // get all instruments since last sync -- add to JSON
        luri = SessionsDB.CONTENT_INSTRUMENTS_URI;
        lprojection = new String[] {SessionsDB._ID, SessionsDB.INSTRUMENT_NAME, SessionsDB.INSTRUMENT_DESC, SessionsDB.INSTRUMENT_DATA_VER};
        lselection = SessionsDB.INSTRUMENT_DATA_VER + " >= " + ldb_last_sync;
		lcursor = mContext.getContentResolver().query(luri, lprojection, lselection, null, null);
		if (null != lcursor) {
			Log.d(LOGTAG, "doInBackground() 43AA   Instruments Q cur: " + lcursor.toString() + " sel: " + lselection + " proj: " + lprojection);
			lrows = lcursor.getCount();
			Log.w(LOGTAG, "doInBackground() 43BB cursor lrows: " + lrows );
			lcursor.moveToFirst();
			
			for (int i = 0;i < lrows; i++) {
			    ljObject = new JSONObject();
			    
			    try {
				    // table
				    ljObject.put(SCRUD_OBJECT, SCRUD_OBJECT_INSTRUMENTS);
				    // operation
				    ljObject.put(SCRUD_OP, SCRUD_OP_INSUPD);
				    // data fields
				    ljObject.put(SessionsDB._ID, lcursor.getInt(0));
				    ljObject.put(SessionsDB.INSTRUMENT_NAME, lcursor.getString(1));
					ljObject.put(SessionsDB.INSTRUMENT_DESC, lcursor.getString(2));
					ljObject.put(SessionsDB.INSTRUMENT_DATA_VER, lcursor.getString(3));
				} catch (JSONException e) {
		            Log.e(LOGTAG, "ERROR: doInBackground() 43XX  JSON Exception INSTRUMENTS -- exiting");
					e.printStackTrace();
				}
				Log.d(LOGTAG, "doInBackground() 43YY   ljobject: " + ljObject.toString());
			    ljArray.put(ljObject);
			    
				lcursor.moveToNext();
			} // END for()
		} else {
			Log.d(LOGTAG, "doInBackground() 44AA   Goals Q cur: " + lcursor + " sel: " + lselection + " proj: " + lprojection);
		}
		Log.d(LOGTAG, "doInBackground() 44ZZ   len: " +  ljArray.length() + " jarrayStr: " + ljArray.toString());	
		
        // get all practices since last sync -- add to JSON
		// FIXME: this really needs to be broken into buckets/chunks... breaking one big xfer into several small ones, each checking return code and updating status...
        luri = SessionsDB.CONTENT_PRACTICES_URI;
        lprojection = new String[] {SessionsDB._ID, SessionsDB.PRACTICE_DATE_STAMP, SessionsDB.PRACTICE_LENGTH_TOTAL, SessionsDB.PRACTICE_LENGTH_NOISE, SessionsDB.PRACTICE_AUDIO_FILE, SessionsDB.PRACTICE_AUDIO_FILE_KB, SessionsDB.PRACTICE_GOAL_ID, SessionsDB.PRACTICE_NOTES, SessionsDB.PRACTICE_INSTRUMENT_ID, SessionsDB.PRACTICE_DATA_VER};
        lselection = SessionsDB.PRACTICE_DATA_VER + " >= " + ldb_last_sync;
		lcursor = mContext.getContentResolver().query(luri, lprojection, lselection, null, SessionsDB.PRACTICE_DATA_VER);
		if (null != lcursor) {
			Log.d(LOGTAG, "doInBackground() 53AA   Practices Q cur: " + lcursor.toString() + " sel: " + lselection + " proj: " + lprojection);
			lrows = lcursor.getCount();
			Log.w(LOGTAG, "doInBackground() 53BB cursor lrows: " + lrows );
			lcursor.moveToFirst();
			
			for (int i = 0;i < lrows; i++) {
			    ljObject = new JSONObject();
			    
			    try {
				    // table
				    ljObject.put(SCRUD_OBJECT, SCRUD_OBJECT_PRACTICES);
				    // operation
				    ljObject.put(SCRUD_OP, SCRUD_OP_INSUPD);
				    // data fields
				    ljObject.put(SessionsDB._ID, lcursor.getInt(0));
				    ljObject.put(SessionsDB.PRACTICE_DATE_STAMP, lcursor.getLong(1));
				    ljObject.put(SessionsDB.PRACTICE_LENGTH_TOTAL, lcursor.getLong(2));
				    ljObject.put(SessionsDB.PRACTICE_LENGTH_NOISE, lcursor.getLong(3));
				    ljObject.put(SessionsDB.PRACTICE_AUDIO_FILE, lcursor.getString(4));
				    ljObject.put(SessionsDB.PRACTICE_AUDIO_FILE_KB, lcursor.getLong(5));
				    ljObject.put(SessionsDB.PRACTICE_GOAL_ID, lcursor.getInt(6));
				    ljObject.put(SessionsDB.PRACTICE_NOTES, lcursor.getString(7));
				    ljObject.put(SessionsDB.PRACTICE_INSTRUMENT_ID, lcursor.getInt(8));
				    ljObject.put(SessionsDB.PRACTICE_DATA_VER, lcursor.getLong(9));
				} catch (JSONException e) {
		            Log.e(LOGTAG, "ERROR: doInBackground() 53XX  JSON Exception PRACTICES -- exiting");
					e.printStackTrace();
				}
				Log.d(LOGTAG, "doInBackground() 53YY   ljobject: " + ljObject.toString());
			    ljArray.put(ljObject);
			    
				lcursor.moveToNext();
			} // END for()
		} else {
			Log.d(LOGTAG, "doInBackground() 54AA   Goals Q cur: " + lcursor + " sel: " + lselection + " proj: " + lprojection);
		}
		
		Log.d(LOGTAG, "doInBackground() 60 JSON packed  len: " +  ljArray.length() + " jarrayStr: " + ljArray.toString());	
		
		
		// proceed filling chunk with practices, sorted by data_ver
        
        // select all 
 
        
//        try {
//    	// credit SO Herly -- http://stackoverflow.com/questions/4910415/android-sending-data-to-be-stored-in-mysql
//    	// Sample upload data
//    	ArrayList<NameValuePair> nameValuePairs = new ArrayList<NameValuePair>(2);
//// FIXME: review and trash or keep...    	
////        nameValuePairs.add(new BasicNameValuePair("Lat","19.80"));
////        nameValuePairs.add(new BasicNameValuePair("Lon","13.22"));
//        nameValuePairs.add(new BasicNameValuePair("Lat","howdy"));
//        nameValuePairs.add(new BasicNameValuePair("Lon","doody"));

        // FIXME: create a 'send_batch()' method
        try {
        	lentity = new StringEntity(ljArray.toString(), HTTP.UTF_8);
        	lentity.setContentType("application/json");
        	////httppost.setHeader("Content-Type","application/json");
        	httppost.setEntity(lentity);

        	HttpResponse response = httpclient.execute(httppost);
            HttpEntity entity = response.getEntity();

            InputStream is = entity.getContent();
            Log.w(LOGTAG, "doInBackground() 84 response:" + response.getStatusLine().toString());
            Log.w(LOGTAG, "doInBackground() 85 is:" + is.toString() ); // FIXME: review and delete --  .read() );
            Log.w(LOGTAG, "doInBackground() 86 is str:" + istreamToStr(is, 3000) ); // FIXME: review and delete --  .read() );

            // Check status, and update local db.
            // FIXME: improve, check return codes/actions for each record.  Handle return data and crud commands.
            StatusLine lstatus = response.getStatusLine();
//            if (200 == lstatus.getStatusCode()) { 
//                Log.w(LOGTAG, "doInBackground() 88  updating sync_date_stamp in sync db");
//                ContentValues lval = new ContentValues();
//                lval.put(SessionsDB.SYNC_DATE_STAMP, lnewNowSec);
//                luri = SessionsDB.CONTENT_SYNC_URI;
//                lselection = SessionsDB._ID + " = 1 ";
//                int lrc = mContext.getContentResolver().update(luri, lval, lselection, null);
//                ldb_userid = lpref_userid;
//                Log.w(LOGTAG, "doInBackground() 89  updating userid sync db RC: " + lrc);
//            }

        
        } catch(Exception e) {
            Log.e(LOGTAG, "ERROR: doInBackground() 93 response in http connection " + e.toString());
            Log.w(LOGTAG, "ERROR: doInBackground() 94 e trace: " + e.getLocalizedMessage());
        }       
        
        
        
        	// FIXME: review and delete -- trash from connection android page -- https://developer.android.com/training/basics/network-ops/connecting.html#AsyncTask
//            URL url = new URL(urlStr);
//            HttpURLConnection conn = (HttpURLConnection) url.openConnection();
//            conn.setReadTimeout(10000 /* milliseconds */);
//            conn.setConnectTimeout(15000 /* milliseconds */);
//            conn.setRequestMethod("POST");
//            conn.setDoInput(true);
//
//            // Starts the query
//            conn.connect();
//            int response = conn.getResponseCode();
//            Log.d(DEBUG_TAG, "The response is: " + response);
//            is = conn.getInputStream();
//
//            // Convert the InputStream into a string
//            String contentAsString = readIt(is, len);
//            return contentAsString;
            
        // Makes sure that the InputStream is closed after the app is
        // finished using it.

        
        
        
        // send basic data
        //		send big data (table) [after query]
        // check result
        
        

        publishProgress(1);
        //not good here -- cancel(true);        
        publishProgress(2);
        Log.d(LOGTAG, "doInBackground() 98 cancelled:" + isCancelled());
        return 0;
	} // END doInBackground()
	

	@Override
    protected void onProgressUpdate(Integer... progress) {
        Log.d(LOGTAG, "onProgressUpdate() 00   i progress: " + progress.toString() + " prog[0]: " + progress[0]);
        // FIXME: are checks for null required? -- probably not 

        if ( (null != progress) && (0 == progress[0]) ) {
            Log.d(LOGTAG, "onProgressUpdate() 11   CANCELLING");
        	cancel(true);        
        }
        //setProgressPercent(progress[0]);
    }
	
	@Override
	protected void onPostExecute(Integer result) {
        Log.d(LOGTAG, "onPostExecute() 00   i result: " + result);
		Toast.makeText(mContext,"Sync completed OK - @post exec of mpq sync task", Toast.LENGTH_SHORT).show();
		
		// Clear out any notifications.
		mNotificationManager.cancel(NOTIFICATION_SYNC_ID);
	}
	

	// FIXME: this does not appear to be called when cancel() is called?!?
	@Override
	protected void onCancelled(Integer result) {
        Log.d(LOGTAG, "onCancelled() 00   i result: " + result);

        // Clear out any notifications.
		mNotificationManager.cancel(NOTIFICATION_SYNC_ID);
	}
	
	
	

    /**********************
     * 
     * Private worker methods for the class, such as notification management.
     */
    
    /**
     * Show a notification.
     * FIXME: this method started as a copy from the same method in MPQAudioService.java *** (? would it benefit to call the original, statically?)
     */
    @SuppressWarnings("deprecation")
	private void showNotification() {
    	CharSequence ltext = mContext.getText(R.string.notification_sservice_syncing);
    	CharSequence ltitle = mContext.getText(R.string.notification_aservice);
    	
        // Set the icon, scrolling text and timestamp
    	NotificationCompat.Builder builder = new NotificationCompat.Builder(mContext);
    	builder.setSmallIcon(R.drawable.ic_launcher);
    	builder.setContentTitle(ltitle);
    	builder.setContentText(ltext);
    	//builder.setAutoCancel(true); // Enable the automatic close of the notification, once the user clicks it.

// FIXME: add a way for the notification to allow the user to cancel sync task...    	
//    	Intent intent = new Intent(this, MPQMainTabActivity.class);
//        // The PendingIntent to launch our activity if the user selects this notification
//        PendingIntent lcontentIntent = PendingIntent.getActivity(this, 0, intent, 0); // FIXME: keep or toss this -- PendingIntent.FLAG_CANCEL_CURRENT);
//    	builder.setContentIntent(lcontentIntent);

        // Build and send the notification and begin running service in the foreground.
        Notification n = builder.build();
        
        PendingIntent lpi = PendingIntent.getActivity(mContext, 0, null, 0);
        n.setLatestEventInfo(mContext, ltitle, ltext, lpi);
      
        if (null == mNotificationManager) {
        	mNotificationManager = (NotificationManager) mContext.getSystemService(Context.NOTIFICATION_SERVICE);
        }
        
        // NOTIFICATION_SYNC_ID allows update of notification later on.
    	mNotificationManager.notify(NOTIFICATION_SYNC_ID, n);
        
    } // END showForegroundNotification()
    

    /**
     * Reads an InputStream and converts it to a String.
     * credit: android training connecting to the network -- https://developer.android.com/training/basics/network-ops/connecting.html#AsyncTask readIt()
     */
    private String istreamToStr(InputStream stream, int len) throws IOException, UnsupportedEncodingException {
        Reader reader = null;
        reader = new InputStreamReader(stream, "UTF-8");        
        char[] buffer = new char[len];
        int iread = reader.read(buffer);
        Log.d(LOGTAG, "istreamToStr() 99  iread: " + iread);
        return new String(buffer);
    }
    
    
    /**
     * Get device mobile network ID (ex: ESN).
     * credit: http://android-coding.blogspot.com/2011/06/get-unique-device-id-imei-meid-or-esn.html
     */
//    //retrieve a reference to an instance of TelephonyManager
//    TelephonyManager telephonyManager = (TelephonyManager)getSystemService(Context.TELEPHONY_SERVICE);
//    textDeviceID.setText(getDeviceID(telephonyManager));
    private String getDeviceID(TelephonyManager phonyManager){
    	String id = phonyManager.getDeviceId();
    	if (id == null){
    		id = "not available";
    	}
    	 
    	int phoneType = phonyManager.getPhoneType();
    	switch(phoneType){
    	case TelephonyManager.PHONE_TYPE_NONE:
    		return "NONE: " + id;
    	case TelephonyManager.PHONE_TYPE_GSM:
    		return "GSM: IMEI=" + id;
   	 	case TelephonyManager.PHONE_TYPE_CDMA:
//   	 		return "CDMA: MEID/ESN=" + id;
   	 		// FIXME: possibly return two variables, one with type, and the other with ID.  For now we have only ESN, and need the id only.
   	 		return id;
    	 /**
    	  *  for API Level 11 or above
    	  *  case TelephonyManager.PHONE_TYPE_SIP:
    	  *   return "SIP";
    	  */
    	default:
    		return "UNKNOWN: ID=" + id;
    	} // END switch
    	 
    } // END getDeviceID()

    
} // END class MPQSyncTask
